--
-- BartStation Video Controller
-- by Bart Trzynadlowski, 2005-2006
--
-- A video controller designed for use with the Spartan-3 Starter Kit Board for
-- the BartStation homebrew video game system project.
--

--
-- vga.vhd
--
-- Implementation of the VGA video controller.
--

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library UNISIM;	-- Xilinx primitives
use UNISIM.VComponents.all;

--
-- VGA
--
-- Defines the VGA controller interface.
--
entity VGA is
	port
	(
		--
		-- 50MHz Clock
		--
		clk_50mhz:	in std_logic;

		--
		-- VGA Signals
		--
		vga_rgb:	out std_logic_vector(2 downto 0);
		vga_hs:	out std_logic;
		vga_vs:	out std_logic;
		vga_vbl:	out std_logic;

		--
		-- SRAM Pins
		--
		-- These are mapped to the actual FPGA pins connected to the SRAM IC.
		--
		sram_pins_addr:out std_logic_vector(17 downto 0);
		sram_pins_io: 	inout std_logic_vector(7 downto 0);
		sram_pin_oe: 	out std_logic;
		sram_pin_we:	out std_logic;
		sram_pin_ce: 	out std_logic;
		sram_pin_ce2: 	out std_logic;
		sram_pin_ub: 	out std_logic;
		sram_pin_lb: 	out std_logic;

		--
		-- Input Port
		--		
		-- input(10):	Control Register Enable. Active low.
		-- input(9):	Video Data Register Enable. Active low.
		-- input(8):	Input Clock. Writes are processed on falling edge.
		-- input(7):	D7
		-- input(6):	D6
		-- input(5):	D5
		-- input(4):	D4
		-- input(3):	D3
		-- input(2):	D2 (R)
		-- input(1):	D1 (G)
		-- input(0):	D0 (B)
		--
		input_pins:	in std_logic_vector(10 downto 0)
	);
end VGA;

--
-- VGA: Behavioral Architecture
--
-- Implementation of the VGA controller.
--
architecture Behavioral of VGA is
	--
	-- VGA Signals
	--
	-- A 25MHz clock is generated for VGA timing.
	--
	signal clk_25mhz: 			std_logic;
	signal horizontal_counter:	std_logic_vector(9 downto 0);
	signal vertical_counter:	std_logic_vector(9 downto 0);
	signal framebuf_addr:		std_logic_vector(18 downto 0);
	signal framebuf_line_addr:	std_logic_vector(18 downto 0);
	signal display_enable:		std_logic;
	signal interrupt_enable:	std_logic := '0';
	
	--
	-- Inport Port Signals
	--
	signal input:					std_logic_vector(10 downto 0);
	signal input_clk:				std_logic := '0';
	signal input_clk_prev:		std_logic := '0';
	
	signal mode:					std_logic_vector(2 downto 0);	
	
	signal addr_x:					std_logic_vector(8 downto 0);
	signal addr_y:					std_logic_vector(7 downto 0);
	signal write_addr:			std_logic_vector(17 downto 0);

	--
	-- SRAM Signals and State Management
	--
	signal sram_addr:		std_logic_vector(17 downto 0);
	signal sram_write_addr:	std_logic_vector(17 downto 0);
	signal sram_read_addr:	std_logic_vector(17 downto 0);
	signal sram_data_write:	std_logic_vector(7 downto 0);
	signal sram_data_read:	std_logic_vector(7 downto 0);
	signal sram_ce: 		std_logic;
	signal sram_ce2:		std_logic;
	signal sram_oe: 		std_logic;
	signal sram_we: 		std_logic;
	signal sram_ub: 		std_logic;
	signal sram_lb: 		std_logic;
	signal sram_io_t:		std_logic;
	
	type sram_state_t is
	(
		SRAM_WAITING,	-- wait for command, place data on bus
		SRAM_WRITE,		-- complete a write transaction
		SRAM_READ		-- complete a read transaction
	);
	signal sram_state:		sram_state_t := SRAM_WAITING;

	signal sram_do_read:		std_logic := '0';
	signal sram_do_write:	std_logic := '0';

	signal sram_data_read_buf:		std_logic_vector(7 downto 0);
	signal sram_data_write_buf:	std_logic_vector(7 downto 0);
begin
  	--
	-- Map SRAM pins to Xilinx IO buffers
	--
	g1: 
		for i in sram_pins_io'range generate
  			iod0: iobuf port map 
  			(
				i => sram_data_write(i),
   			io => sram_pins_io(i),
   			o => sram_data_read(i),
   			t => sram_io_t	  -- 0 read?
			);
		end generate g1; 
	g2: 
		for i in sram_pins_addr'range generate
  			oa0: obuf port map
  			(
				o => sram_pins_addr(i),
   			i => sram_addr(i)
			);
		end generate g2;
	we: obuf port map
  		(
			o => sram_pin_we,
   		i => sram_we
		);

	oe: obuf port map
  		(
			o => sram_pin_oe,
   		i => sram_oe
		);
	ce: obuf port map
  		(
			o => sram_pin_ce,
   		i => sram_ce
		);
	ce2: obuf port map
  	(
		o => sram_pin_ce2,
   	i => sram_ce2
	);
	ub: obuf port map
  		(
			o => sram_pin_ub,
   		i => sram_ub
		);
	lb: obuf port map
  		(
			o => sram_pin_lb,
   		i => sram_lb
		);
	
	inp:
		for i in input_pins'range generate
  		np0: ibuf port map
  			(
				i => input_pins(i),
   			o => input(i)
			);
		end generate inp;
	
	--
	-- Fixed signals
	--
	sram_ce2 <= '1';	-- make sure second SRAM chip stays off the bus
	sram_lb <= '0';
	sram_ub <= '1';

	--
	-- Generate the 25MHz clock
	--
	process(clk_50mhz)
	begin
		if clk_50mhz'event and clk_50mhz = '1' then
			clk_25mhz <= (clk_25mhz xor '1');
		end if;
	end process;

	--
	-- Input Port State Machine
	--
	process(clk_50mhz)
	begin
		if clk_50mhz'event and clk_50mhz = '1' then
			input_clk <= input(8);			-- latch the asynchronous input clock
			input_clk_prev <= input_clk;	-- save the previous latched clock value

		
			if input_clk_prev = '1' and input_clk = '0' then	-- falling edge
				if input(10) = '0' then						-- Control Register enabled
					mode <= input(2 downto 0);				-- mode bits				
				elsif input(9) = '0' then					-- Video Data Register enabled
					case mode is
					when "000" =>	-- data write mode
						sram_data_write_buf(2 downto 0) <= input(2 downto 0);
						sram_write_addr <= write_addr;
						write_addr <= write_addr + 1;
						sram_do_write <= '1';
					when "001" => 	-- address X low
						addr_x <= addr_x(8) & input(7 downto 0);					
						write_addr <= ("00" & addr_y & "00000000") + ("0000" & addr_y & "000000") + (addr_x(8) & input(7 downto 0));
					when "010" =>	-- address X high
						addr_x(8) <= input(0);
						write_addr <= ("00" & addr_y & "00000000") + ("0000" & addr_y & "000000") + (input(0) & addr_x(7 downto 0));
					when "011" =>	-- address Y
						addr_y <= input(7 downto 0);
						write_addr <= ("00" & input(7 downto 0) & "00000000") + ("0000" & input(7 downto 0) & "000000") + addr_x;
					when "100" =>	-- settings
						display_enable <= input(0);
						interrupt_enable <= input(1);
					when others =>
					end case;				
				end if;
			end if;		
		
			--
			-- When a write has completed, we must lower the write signal
			--
			if sram_state = SRAM_WRITE then
				sram_do_write <= '0';
			end if;					
		end if;
	end process;


	--
	-- SRAM State Machine
	--
	process(clk_50mhz)
	begin
		sram_oe <= '0';

		if clk_50mhz'event and clk_50mhz = '1' then
			case sram_state is
			when SRAM_WAITING =>
				if sram_do_write = '1' then	-- write requested
					sram_addr <= sram_write_addr;
					sram_data_write <= sram_data_write_buf;
					sram_ce <= '0';
					sram_we <= '0';
					sram_io_t <= '0';										
					sram_state <= SRAM_WRITE;
				elsif sram_do_read = '1' then	-- read requested				
					sram_addr <= sram_read_addr;
					sram_ce <= '0';												
					sram_we <= '1';
					sram_io_t <= '1';					
					sram_state <= SRAM_READ;				
				end if;
			when SRAM_READ =>
				sram_ce <= '1';											
				sram_we <= '1';
				sram_io_t <= '0';
				sram_data_read_buf <= sram_data_read;
				sram_state <= SRAM_WAITING;
			when SRAM_WRITE =>
				sram_ce <= '1';
				sram_we <= '1';
				sram_io_t <= '0';
				sram_state <= SRAM_WAITING;		
			end case;
		end if;
	end process;

	--
	-- VGA Output
	--
	process(clk_25mhz)
	begin
		if clk_25mhz'event and clk_25mhz = '1' then			
			--
			-- When starting a new line, save the frame buffer address and only
			-- update it every other line to halve the vertical resolution. Note
			-- that the display begins at V = 39 so therefore, every ODD line 
			-- will have the frame buffer updated and saved and the even line will
			-- restart there
			--
			if (horizontal_counter = "0000000000") then	
				if (vertical_counter >= "0000000000") and (vertical_counter <= "0000100111") then	-- 39
					framebuf_addr <= "0000000000000000000";
					framebuf_line_addr <= "0000000000000000000";
				else
					--
					-- On odd lines, advance frame buffer by 320 pixels
					--
					if vertical_counter(0) = '1' then	-- odd lines
						framebuf_line_addr <= framebuf_line_addr + "0000000000101000000";
						framebuf_addr <= framebuf_line_addr + "0000000000101000000";
					else
						framebuf_addr <= framebuf_line_addr;
					end if;
				end if;
			end if;


			--
			-- If the horizontal counter is between 144 and 783, we are within the
			-- 640 pixel range of the visible part of the scanline. If the
			-- vertical counter is between 39 and 518, we are in the visible part
			-- of the screen (480 lines.)
			--
			if (horizontal_counter >= "0010010000") 		-- 144
    			and (horizontal_counter < "1100010000")		-- 784
    			and (vertical_counter >= "0000100111") 		-- 39
    			and (vertical_counter < "1000000111") then	-- 519

				--
				-- Fetch the current pixel from SRAM (2 50MHz clock cycles) and
				-- display it.
				--
				
				if display_enable = '1' then
					sram_read_addr <= framebuf_addr(17 downto 0);
					sram_do_read <= '1';
					vga_rgb <= sram_data_read_buf(2 downto 0);
				else
					sram_do_read <= '0';
					vga_rgb <= "000";
				end if;
				
				--
				-- Only advance every other pixel to halve the resolution
				--
				if horizontal_counter(0) = '1' then
					framebuf_addr <= framebuf_addr + 1;
				end if;
			else
				--
				-- We are outside of the visible region, output black
				--				
				vga_rgb <= "000";
				sram_do_read <= '0';
			end if;

			if (vertical_counter < "0000100111") or (vertical_counter >= "1000000111") then
				if interrupt_enable = '1' then
					vga_vbl <= '0';	-- active low
				else
					vga_vbl <= '1';
				end if;
			else
				vga_vbl <= '1';
			end if;

			--
			-- Within this horizontal range, the H-sync signal must be low
			--
			if (horizontal_counter > "0000000000")
      		and (horizontal_counter < "0001100001") then	-- 96+1
				vga_hs <= '0';
    		else
      		vga_hs <= '1';
    		end if;

			--
			-- Generate the V-sync signal in the appropriate range
			--
			if (vertical_counter > "0000000000")
				and (vertical_counter < "0000000011") then	-- 2+1
    			vga_vs <= '0';
    		else
      		vga_vs <= '1';
    		end if;

			--
			-- Increment the horizontal and vertical counters and reset them when 
			-- starting a new line.
			--
    		horizontal_counter <= horizontal_counter+1;
    		if (horizontal_counter = "1100100000") then	-- 800
      		vertical_counter <= vertical_counter+1;
      		horizontal_counter <= "0000000000";
    		end if;
    	
			if (vertical_counter="1000001001") then		-- 521
      		vertical_counter <= "0000000000";
    		end if;

		end if;
	end process;

end Behavioral;